package be.iminds.ilabt.jfed.espec.bundle;

import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.parser.ExperimentSpecificationParser;
import be.iminds.ilabt.jfed.util.IOUtils;
import org.apache.commons.compress.archivers.ArchiveEntry;
import org.apache.commons.compress.archivers.ArchiveException;
import org.apache.commons.compress.archivers.ArchiveInputStream;
import org.apache.commons.compress.archivers.ArchiveStreamFactory;
import org.apache.commons.compress.compressors.CompressorException;
import org.apache.commons.compress.compressors.CompressorInputStream;
import org.apache.commons.compress.compressors.CompressorStreamFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.function.Supplier;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * Bundle
 */
public class AbstractArchiveESpecBundle implements ESpecBundle {
    private static final Logger LOG = LoggerFactory.getLogger(AbstractArchiveESpecBundle.class);

    private final byte[] bytes;
    @Nonnull private final InputStreamSupplierThrowingIOException inputStreamSupplier;

    @Nonnull private final List<ArchiveEntry> archiveEntries;
    @Nonnull private final String eSpecYmlContent;

    private final boolean isCompressedArchive;

    @Nonnull private final CompressorStreamFactory compressorStreamFactory;
    @Nonnull private final ArchiveStreamFactory archiveStreamFactory;


    @Nonnull private final ExperimentSpecification eSpec;


    public AbstractArchiveESpecBundle(InputStream singleUseInputStream) throws ESpecBundleInitException {
        this(() -> singleUseInputStream, true);
    }

    public interface InputStreamSupplierThrowingIOException {
        InputStream get() throws IOException;
    }

    /**
     * @param inputStreamSupplier
     * @param onlyOnce if true, the stream will be fetched and used only once, and the content will be buffered in memory. If false, the content will not be buffered, and the stream will be fetched and read multiple times.
     * @throws ESpecBundleInitException
     */
    AbstractArchiveESpecBundle(@Nonnull InputStreamSupplierThrowingIOException inputStreamSupplier, boolean onlyOnce) throws ESpecBundleInitException {
        if (onlyOnce) {
            try {
                this.bytes = IOUtils.streamToByteArray(inputStreamSupplier.get());
            } catch (IOException e) {
                throw new ESpecBundleInitException("Error reading bundle from InputStream", e);
            }
        } else {
            this.bytes = null;
        }

        this.inputStreamSupplier = onlyOnce ?
                () -> new ByteArrayInputStream(bytes) :
                inputStreamSupplier;

        this.archiveEntries = new ArrayList<>();
        String foundESpecYmlContent = null;

//        boolean isCompressedArchive = archiveFile.getName().endsWith(".tar.gz") || archiveFile.getName().endsWith(".tar.bz2") || archiveFile.getName().endsWith(".tgz");

        compressorStreamFactory = new CompressorStreamFactory();
        archiveStreamFactory = new ArchiveStreamFactory();

        //peek to check if this is a compressed archive, like .tar.gz or .tar.bz2 or .tgz
        boolean isCompressedArchive;
        try (InputStream inputStream = inputStreamSupplier.get();
             BufferedInputStream bufferedInputStream = new BufferedInputStream(inputStream);
             CompressorInputStream cis = compressorStreamFactory.createCompressorInputStream(bufferedInputStream)
        ) {
            isCompressedArchive = true;
        } catch (CompressorException | IOException e) {
            isCompressedArchive = false;
        }
        this.isCompressedArchive = isCompressedArchive;
        LOG.debug("Inputstream was "+(isCompressedArchive?"a":"NOT a")+" compressed archive");

        List<String> foundFilenames = new ArrayList<>();

        try (InputStream inputStream = inputStreamSupplier.get();
             BufferedInputStream bufferedInputStream = new BufferedInputStream(inputStream);
             ArchiveInputStream archiveInput = isCompressedArchive ?
                     archiveStreamFactory.createArchiveInputStream(
                             new BufferedInputStream(compressorStreamFactory.createCompressorInputStream(bufferedInputStream))
                     ) :
                     archiveStreamFactory.createArchiveInputStream(bufferedInputStream)
        ) {
            ArchiveEntry archiveEntry = archiveInput.getNextEntry();
            while (archiveEntry != null) {
                archiveEntries.add(archiveEntry);

                foundFilenames.add(archiveEntry.getName());
                if (archiveEntry.getName().equals(ESPEC_DESCRIPTION_FILENAME)) {
                    byte[] content = new byte[(int) archiveEntry.getSize()];
                    int offset = 0;
                    while(offset < archiveEntry.getSize()) {
                        int read = archiveInput.read(content, offset, content.length - offset);
                        if (read == -1)
                            throw new IOException("Unexpected end of stream");
                        offset += read;
                        assert offset <= archiveEntry.getSize() : "Read to much from stream. This is unexpected. " +
                                "read="+read+" offset="+offset+" archiveEntry.getSize()="+archiveEntry.getSize();
                    }
                    foundESpecYmlContent = new String(content, StandardCharsets.UTF_8);
                }

                archiveEntry = archiveInput.getNextEntry();
            }
            archiveInput.close();
        } catch (IOException | CompressorException | ArchiveException e) {
            throw new ESpecBundleInitException("Failed to process ESpec archive", e);
        }

        if (foundESpecYmlContent != null) {
            this.eSpecYmlContent = foundESpecYmlContent;
        } else {
            throw new RuntimeException("Archive does not contain the required file \""+ ESPEC_DESCRIPTION_FILENAME +"\". Files in archive: "+foundFilenames);
        }

        ExperimentSpecificationParser parser = new ExperimentSpecificationParser();
        try {
            eSpec = parser.parse(this.getExperimentSpecificationYml());
        } catch (ExperimentSpecificationParser.ExperimentSpecificationParseException e) {
            throw new IllegalArgumentException("Failed to parse ExperimentSpecification", e);
        } catch (IOException e) {
            throw new IllegalArgumentException("Failed to read ExperimentSpecification", e);
        }
    }

    private Pattern inDir = Pattern.compile(".*/.+$");

    @Nonnull
    @Override
    public List<String> getFileNames() {
        return archiveEntries.stream()
                .map(ArchiveEntry::getName)
                .filter(name -> !inDir.matcher(name).matches())
                .collect(Collectors.toList());
    }

    @Nonnull
    public byte[] getFileContent(@Nonnull String fileName) throws IOException {

        try (InputStream inputStream = inputStreamSupplier.get();
             BufferedInputStream bufferedInputStream = new BufferedInputStream(inputStream);
             ArchiveInputStream archiveInput = isCompressedArchive ?
                     archiveStreamFactory.createArchiveInputStream(new BufferedInputStream(compressorStreamFactory.createCompressorInputStream(bufferedInputStream))) :
                     archiveStreamFactory.createArchiveInputStream(bufferedInputStream)
        ) {
            ArchiveEntry archiveEntry = archiveInput.getNextEntry();
            while (archiveEntry != null) {
                if (archiveEntry.getName().equals(fileName)) {
                    byte[] content = new byte[(int) archiveEntry.getSize()];
                    int offset = 0;
                    while(offset < archiveEntry.getSize()) {
                        int read = archiveInput.read(content, offset, content.length - offset);
                        if (read == -1)
                            throw new IOException("Unexpected end of stream");
                        offset += read;
                        assert offset <= archiveEntry.getSize() : "Read to much from stream. This is unexpected. " +
                                "read="+read+" offset="+offset+" archiveEntry.getSize()="+archiveEntry.getSize();
                    }
                    return content;
                }

                archiveEntry = archiveInput.getNextEntry();
            }
            archiveInput.close();
        } catch (ArchiveException | CompressorException e) {
            throw new IOException("Failed to process ESpec archive", e);
        }

        throw new IllegalArgumentException("There is no file with name \""+fileName+"\"");
    }

    @Override
    public boolean isDir(@Nonnull String filename) throws IOException {
        String dirFilenameSlash = assureEndWithSlash(filename);
        return archiveEntries.stream()
                .filter(Objects::nonNull)
                .filter(e -> Objects.equals(assureEndWithSlash(e.getName()), dirFilenameSlash))
                .findAny()
                .orElseThrow(() -> new FileNotFoundException("There is no file with name \""+filename+"\""))
                .isDirectory();
    }

    @Override
    public List<String> getDirFiles(@Nonnull String dirFileName) throws IOException {
        if (!isDir(dirFileName))
            throw new IllegalArgumentException("file \""+dirFileName+"\" in bundle is not a directory");
        String dirFilenameSlash = assureEndWithSlash(dirFileName);
        return archiveEntries.stream()
                .filter(Objects::nonNull)
                .map(ArchiveEntry::getName)
                .filter(n -> n.length() > dirFilenameSlash.length())
                .filter(n -> n.startsWith(dirFilenameSlash))
                .filter(n -> {
                    int slash = n.indexOf('/', dirFilenameSlash.length());
                    return slash == -1 || slash == n.length()-1;
                })
                .collect(Collectors.toList());
    }

    @Nonnull
    @Override
    public ExperimentSpecification getExperimentSpecification() {
        return eSpec;
    }




    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }
}
